mod bluetooth;
mod patterns;

use crate::patterns::{ColorWipe, Fade, PixelColor, Rainbow};
use bluer::gatt::remote::Characteristic;
use patterns::Strip;
use std::convert::TryFrom;
use std::time::Duration;
use tokio::sync::watch;

const SERVICE_UUID: bluer::Uuid = bluer::Uuid::from_u128(0xadaf0900c33242a893bd25e905756cb8);
const PIXEL_DATA_UUID: bluer::Uuid = bluer::Uuid::from_u128(0xadaf0903c33242a893bd25e905756cb8);
const STRIP_SIZE: usize = 25;
const BASE_STRIP_DATA: [u8; 3] = [0x00, 0x00, 0x01];

#[tokio::main(flavor = "current_thread")]
async fn main() -> bluer::Result<()> {
    let adapter = bluetooth::create_session().await?;

    let devices = bluetooth::scan_devices(adapter, SERVICE_UUID).await;

    let mut pixels = Vec::<PixelColor>::new();

    for _ in 0..STRIP_SIZE {
        pixels.push(PixelColor {
            red: 255,
            green: 0,
            blue: 0,
        })
    }

    // let strip = Strip::<STRIP_SIZE> {
    //     strip: <[PixelColor; 25]>::try_from(pixels).unwrap(),
    // };

    let (tx, rx) = watch::channel(Strip::<STRIP_SIZE>::default());

    for device in devices {
        bluetooth::connect(&device, 3).await?;
        let char = bluetooth::get_char(&device, SERVICE_UUID, PIXEL_DATA_UUID).await?;
        if let Some(char) = char {
            let mut rx_device = rx.clone();
            tokio::spawn(async move {
                println!("device connected : {}", &device.address());
                while rx_device.changed().await.is_ok() {
                    let strip = *rx_device.borrow();
                    write_strip(&strip, &char).await.unwrap();
                }
            });
        }
    }

    let mut colorwipe = ColorWipe::<25>{
        current_iteration: 0,
        color: PixelColor {
            red: 0,
            green: 255,
            blue: 0
        },
        infinite: true
    };


    let mut fade = Fade::<25> {
        current_iteration: 0,
        nbr_iterations: 25,
        begin_color: PixelColor {
            red: 255,
            green: 0,
            blue: 0
        },
        end_color: PixelColor {
            red: 0,
            green: 0,
            blue: 255
        },
    };

    let mut pattern = Rainbow::<25>{
        current_iteration: 0,
        max_iteration: None,
        step: Some(10),
    };

    let mut strip = fade.next().unwrap();

    while tx.send(strip).is_ok()  {
        if let Some(value) = fade.next() {
            strip = value;
        } else {
            break;
        }
        tokio::time::sleep(Duration::from_millis(200)).await;
    }

    while tx.send(pattern.next().unwrap()).is_ok()  {
        tokio::time::sleep(Duration::from_millis(10)).await;
    }
    Ok(())
}

// fn create_pattern_list() -> Vec<patterns>
// async fn send_seq(char: &Characteristic) -> bluer::Result<()> {
//     println!("  Characteristic flags : {:?}, ", char.flags().await?);
//
//     let mut strip_red : Strip<10> = Strip::default();
//     strip_red.strip.fill(PixelColor { red: 255, green: 0, blue: 0});
//
//     let mut strip_green: Strip<10> = Strip::default();
//     strip_green.strip.fill(PixelColor { red: 0, green: 255, blue: 0});
//
//     let pattern = RainbowPattern::<STRIP_SIZE> { current_iteration: 0,};
//
//
//     for strip in pattern {
//         write_strip(strip, char).await?;
//         sleep(Duration::from_millis(50)).await;
//     }
//
//     Ok(())
// }

pub async fn write_strip<const N: usize>(
    data: &Strip<N>,
    char: &Characteristic,
) -> bluer::Result<()> {
    let frame = [BASE_STRIP_DATA.to_vec(), data.to_array()].concat();
    // print!("{:?}", frame);
    char.write(&*frame).await?;
    Ok(())
}

// async fn find_neopixel_service(device: &Device) -> bluer::Result<Option<Characteristic>> {
//     let addr = device.address();
//     let uuids = device.uuids().await?.unwrap_or_default();
//
//     if uuids.contains(&SERVICE_UUID) {
//         println!("service neopixel found for device {}", &addr);
//         match connect(device, 3).await {
//             Ok(()) => {
//                 println!("successefully connected");
//             }
//             Err(err) => {
//                 println!("unable to connect, trying new device, {}", err);
//                 return Err(err)
//             }
//         }
//         for service in device.services().await? {
//             if SERVICE_UUID == service.uuid().await? {
//                 for char in service.characteristics().await? {
//                     println!("uuid : {}", &char.uuid().await?);
//                     if PIXEL_DATA_UUID == char.uuid().await? {
//
//                         return Ok(Some(char));
//                     }
//
//                 }
//             }
//         }
//
//     }
//
//     Ok(None)
// }
